from sqlalchemy.orm.dynamic import AppenderMixin
from typing import Union


class Field(object):
    """
    Configure a ModelSerializer field
    """

    def __init__(self, dump_only=False, load_only=False, serializer=None):
        self.dump_only = dump_only
        self.load_only = load_only
        self._serializer = serializer

    @property
    def serializer(self):
        return self._serializer


    def dump(self, value):
        if value and self.serializer:
            return self.serializer.dump(value)
        else:
            return value


    def load(self, serialized):
        if serialized and self.serializer:
            return self.serializer.load(serialized)
        else:
            return serialized


class NestedModelField(Field):
    """
    A field to Dump and Update nested models.
    """

    def __init__(self, declarative_class, **kw):
        from .modelserializer import ModelSerializer

        super().__init__(**kw)
        if self._serializer is None:
            self._serializer = ModelSerializer(declarative_class)

    def load(self, serialized):
        if not serialized:
            return None
        class_mapper = self.serializer.model_class
        pk_attr = self._get_pk_attr_name(class_mapper)
        pk = serialized.get(pk_attr)
        if pk:
            # Serialized object has a primary key, so we load an existing model from the database
            # instead of creating one
            existing_model = class_mapper.query.get(pk)
            return self.serializer.load(serialized, existing_model)
        else:
            # No primary key, just create a new model entity
            return self.serializer.load(serialized)

    @staticmethod
    def _get_pk_attr_name(declarative_model):
        primary_keys = declarative_model.__mapper__.primary_key
        assert len(primary_keys) == 1, "Nested object must have exactly one primary key"
        pk_name = primary_keys[0].key
        return pk_name


class NestedAttributesField(Field):
    """
    A read-only field that dump nested object attributes.
    """
    from .serializer import Serializer

    class NestedAttributesSerializer(Serializer):

        def __init__(self, attributes, many):
            self.attributes = attributes
            self.many = many

        def dump(self, value):
            if self.many:
                serialized = [self._dump_item(item) for item in value]
            else:
                return self._dump_item(value)
            return serialized

        def _dump_item(self, item):
            serialized = {}
            for attr_name in self.attributes:
                serialized[attr_name] = getattr(item, attr_name)
            return serialized

        def load(self, serialized):
            raise NotImplementedError()


    def __init__(self, attributes: Union[tuple, dict], many=False):
        serializer = self.NestedAttributesSerializer(attributes, many)
        super().__init__(dump_only=True, serializer=serializer)


class PrimaryKeyField(Field):
    """
    Convert relationships in a list of primary keys (for serialization and deserialization).
    """
    from .serializer import Serializer

    class PrimaryKeySerializer(Serializer):

        def __init__(self, declarative_class):
            self.declarative_class = declarative_class
            self._pk_column = get_model_pk(self.declarative_class)

        def load(self, serialized):
            pk_column = self._pk_column
            return self.declarative_class.query.filter(pk_column.in_(serialized)).all()

        def dump(self, value):
            pk_column = self._pk_column
            if is_tomany_attribute(value):
                serialized = [getattr(item, pk_column.key) for item in value]
            else:
                return getattr(value, pk_column.key)
            return serialized

    def __init__(self, declarative_class, **kw):
        super().__init__(serializer=self.PrimaryKeySerializer(declarative_class), **kw)


def get_model_pk(declarative_class):
    """
    Get the primary key Column object from a Declarative model class

    :param Type[DeclarativeMeta] declarative_class: a Declarative class

    :rtype: Column
    """
    primary_keys = declarative_class.__mapper__.primary_key
    assert len(primary_keys) == 1, "Nested object must have exactly one primary key"
    return primary_keys[0]


def is_tomany_attribute(value):
    """
    Check if the Declarative relationship attribute represents a to-many relationship.

    :param value: a SQLAlchemy Declarative class relationship attribute

    :rtype: bool
    """
    return isinstance(value, (list, AppenderMixin))


from .serializer import ColumnSerializer


class EnumSerializer(ColumnSerializer):

    def dump(self, value):
        if not value:
            return None
        return value.value

    def load(self, serialized):
        enum = getattr(self.column.type, 'enum_class')
        return enum(serialized)


def is_enum_field(col):
    return hasattr(col.type, 'enum_class') and getattr(col.type, 'enum_class')


from .fields import Field
from .serializer import Serializer


class ModelSerializer(Serializer):
    """
    Serializer for SQLAlchemy Declarative classes
    """

    def __init__(self, model_class):
        """
        :param Type[DeclarativeMeta] model_class: the SQLAlchemy mapping class to be serialized
        """
        self._mapper_class = model_class
        self._fields = self._get_declared_fields()
        # Collect columns not declared in the serializer
        for column in self.model_columns.keys():
            field = self._fields.setdefault(column, Field())
            # Set a serializer for fields that can not be serialized by default
            if field.serializer is None:
                from flask_restalchemy import Api
                serializer = Api.find_column_serializer(self.model_columns.get(column))
                field._serializer = serializer

    @property
    def model_class(self):
        return self._mapper_class

    @property
    def model_columns(self):
        return self._mapper_class.__mapper__.c

    def dump(self, model):
        """
        Create a serialized dict from a Declarative model

        :param DeclarativeMeta model: the model to be serialized

        :rtype: dict
        """
        serial = {}
        for attr, field in self._fields.items():
            if field.load_only:
                continue
            value = getattr(model, attr) if hasattr(model, attr) else None
            if field:
                serialized = field.dump(value)
            else:
                serialized = value
            serial[attr] = serialized
        return serial

    def load(self, serialized, existing_model=None):
        """
        Instancialize a Declarative model from a serialized dict

        :param dict serialized: the serialized object.

        :param None|DeclarativeMeta existing_model: If given, the model will be updated with the serialized data.

        :rtype: DeclarativeMeta
        """
        if existing_model:
            model = existing_model
        else:
            model = self._mapper_class()
        for field_name, value in serialized.items():
            field = self._fields[field_name]
            if field.dump_only:
                continue
            deserialized = field.load(value)
            setattr(model, field_name, deserialized)
        return model

    def get_model_name(self) -> str:
        return self._mapper_class.__name__

    def before_put_commit(self, model, session):
        pass

    def after_put_commit(self, model, session):
        pass

    def before_post_commit(self, model, session):
        pass

    def after_post_commit(self, model, session):
        pass

    @classmethod
    def _get_declared_fields(cls) -> dict:
        fields = {}
        # Fields should be only defined ModelSerializer subclasses,
        if cls is ModelSerializer:
            return fields
        for attr_name in dir(cls):
            if attr_name.startswith('_'):
                continue
            value = getattr(cls, attr_name)
            if isinstance(value, Field):
                fields[attr_name] = value
        return fields



from abc import ABC, abstractmethod


class Serializer(ABC):

    @abstractmethod
    def dump(self, value): pass

    @abstractmethod
    def load(self, serialized): pass


class ColumnSerializer(Serializer):

    def __init__(self, column):
        self.column = column
